import logging
import os
from typing import Any, Dict, Optional, Tuple

import mlflow
import torch
from pytorch_lightning import LightningModule

import sfu_torch_lib.io as io
import sfu_torch_lib.utils as utils

logger = logging.getLogger(__name__)


def get_checkpoint_path(run_id: str, filename: str = 'last') -> Optional[str]:
    run = mlflow.get_run(run_id)

    assert isinstance(run.info.artifact_uri, str)

    checkpoint_path = os.path.join(run.info.artifact_uri, f'{filename}.ckpt')

    if not io.exists(checkpoint_path):
        return None

    return checkpoint_path


def get_localized_checkpoint_path(
    run_id: str,
    filename: str = 'last',
    overwrite: bool = True,
    cache: bool = False,
) -> Optional[str]:
    checkpoint_path = get_checkpoint_path(run_id, filename)

    if checkpoint_path is None:
        return None

    basename = f'{run_id}.ckpt'
    checkpoint_path = io.localize_cached(checkpoint_path, basename, overwrite, cache)

    return checkpoint_path


def get_resumable_checkpoint_path(
    run_id: Optional[str],
    run_id_pretrained: Optional[str],
    filename: str = 'last',
    overwrite: bool = True,
    cache: bool = False,
) -> Tuple[Optional[str], bool]:
    if run_id:
        checkpoint_path = get_localized_checkpoint_path(run_id, filename, overwrite, cache)

        if checkpoint_path:
            return checkpoint_path, False

    if run_id_pretrained:
        checkpoint_path = get_localized_checkpoint_path(run_id_pretrained, filename, overwrite, cache)

        if checkpoint_path:
            return checkpoint_path, True

    return None, True


def get_checkpoint(run_id: str, filename: str = 'last') -> Optional[Dict[str, Any]]:
    checkpoint_path = get_checkpoint_path(run_id, filename)

    if checkpoint_path is None:
        return None

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file)

    return checkpoint


def load_model(
    run_id: str,
    module_class: type[LightningModule] | None = None,
    filename: str = 'last',
    cache: bool = False,
    overwrite: bool = True,
    **kwargs,
) -> LightningModule:
    if cache:
        checkpoint_path = get_localized_checkpoint_path(run_id, filename, overwrite)
    else:
        checkpoint_path = get_checkpoint_path(run_id, filename)

    assert checkpoint_path

    module_class = module_class or utils.get_run_class(run_id)

    with io.open(checkpoint_path) as checkpoint_file:
        model = module_class.load_from_checkpoint(checkpoint_file, **kwargs)

    assert isinstance(model, LightningModule)

    return model


def load_checkpoint_state(checkpoint_path: str, model: LightningModule, strict: bool = True) -> None:
    device = torch.device('cuda') if torch.cuda.device_count() else torch.device('cpu')

    with io.open(checkpoint_path) as checkpoint_file:
        checkpoint = torch.load(checkpoint_file, device)

    model.load_state_dict(checkpoint['state_dict'], strict)


def load_run_state(run_id: str, model: LightningModule, filename: str = 'last', strict: bool = True) -> None:
    checkpoint_path = get_localized_checkpoint_path(run_id, filename)

    assert checkpoint_path

    load_checkpoint_state(checkpoint_path, model, strict)



import functools
import importlib
import re
from types import ModuleType
from typing import List, Tuple, TypeVar, Iterable, Iterator, Type, Union, Sequence, Mapping, Callable, Generic, Dict
from typing import Optional

import mlflow
from pytorch_lightning import LightningModule
from torch.nn import ELU, Identity, ReLU, LeakyReLU, Sigmoid, Softmax, Tanh, Module


T = TypeVar('T')


class AttributeDict(Generic[T]):
    def __init__(self, dictionary: Dict[str, T]) -> None:
        self.dictionary = dictionary

        for key, value in self.items():
            setattr(self, key.lower().replace(' ', '_'), value)

    def __getattr__(self, name):
        return getattr(self, name)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __repr__(self):
        return repr(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __contains__(self, item):
        return item in self.dictionary

    def __iter__(self):
        return iter(self.dictionary)

    def __or__(self, other: 'AttributeDict[T]') -> 'AttributeDict[T]':
        return AttributeDict(self.dictionary | other.dictionary)

    def update(self, other: 'AttributeDict[T]') -> 'AttributeDict[T]':
        return self | other

    def keys(self):
        return self.dictionary.keys()

    def values(self):
        return self.dictionary.values()

    def items(self):
        return self.dictionary.items()


def get_pairs(items: Iterable[T]) -> Iterator[Tuple[T, T]]:
    first = None

    for index, item in enumerate(items):
        if index % 2 == 0:
            first = item
        else:
            yield first, item  # type: ignore


def create_selector(indices: Union[int, Sequence[int]]) -> Callable[[Sequence[T]], Union[T, Tuple[T, ...]]]:
    def select(items: Sequence[T]) -> Union[T, Tuple[T, ...]]:
        if isinstance(indices, int):
            return items[indices]

        return tuple(items[index] for index in indices)

    return select


def to_list(*args: Union[Sequence[T], T]) -> List[T]:
    items: List[T] = []

    for argument in args:
        if isinstance(argument, Sequence):
            items += argument
        else:
            items.append(argument)

    return items


def to_bool(value: Optional[str]) -> Optional[bool]:
    if value is None:
        return None

    if value == '0' or value.lower() == 'false':
        return False

    return bool(value)


def to_bool_or_false(value: Optional[str]) -> bool:
    return to_bool(value) or False


def get_activation(name: str) -> Type[Module]:
    return {
        'elu': ELU,
        'identity': Identity,
        'relu': ReLU,
        'leaky_relu': LeakyReLU,
        'sigmoid': Sigmoid,
        'softmax': Softmax,
        'tanh': Tanh,
    }[name]


def get_class(model_type: str, default_module: Optional[ModuleType] = None) -> Type:
    if '.' in model_type:
        module_name, class_name = model_type.rsplit('.', maxsplit=1)
        module = importlib.import_module(module_name)

    elif default_module is not None:
        module = default_module
        class_name = model_type

    else:
        raise ValueError('default_module not provided')

    model_class = getattr(module, class_name)

    return model_class


def get_run_class(
    run_id: str,
    default_module: Optional[ModuleType] = None,
    key: str = 'model-type',
) -> Type[LightningModule]:
    model_type = mlflow.get_run(run_id).data.params[key]

    model_class = get_class(model_type, default_module)

    return model_class


def get_type_or_run_class(
    model_type: Optional[str] = None,
    run_id: Optional[str] = None,
    default_module: Optional[ModuleType] = None,
    key: str = 'model-type',
) -> Type[LightningModule]:
    if model_type is not None:
        model_class = get_class(model_type, default_module)

    else:
        assert run_id is not None
        model_class = get_run_class(run_id, default_module, key)

    return model_class


def parse_map_sequence(string: str, sequence_type: Type[T]) -> Mapping[str, Sequence[T]]:
    substrings = re.split(r'(?<=]),', string)

    output = {}

    for substring in substrings:
        key, value = substring.split(':')

        output[key] = list(map(sequence_type, value.strip('[]').split(',')))

    return output


parse_map_ints = functools.partial(parse_map_sequence, sequence_type=int)
parse_map_floats = functools.partial(parse_map_sequence, sequence_type=float)
parse_map_strings = functools.partial(parse_map_sequence, sequence_type=str)
